package com.sportstracker.apiserver.integration.fit.examples;

import com.garmin.fit.*;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.StringJoiner;

/**
 * Example code at https://apizone.suunto.com/ on how to decode Suunto specific 'developer fields' from a FIT file.
 *
 * Usage:
 * {@code
 *     byte[] bytes = Files.readAllBytes(Paths.get("c:/workout.fit"));
 *     SuuntoDeveloperFieldsDecodingExample.decode(bytes, System.out);
 * }
 */
public class SuuntoDeveloperFieldsDecodingExample {

    private static class ApplicationId {
        // Application id for fields generated by Suunto watches
        public static final byte[] SUUNTO_STANDARD = { 83, 117, 117, 110, 116, 111, 70, 105, 116, 69, 120, 112, 111, 114, 116, 49 };

        // Application id for Suuntoplus related fields
        public static final byte[] SUUNTOPLUS_METADATA = { 83, 117, 117, 110, 116, 111, 112, 108, 117, 115, 70, 105, 116, 69, 120, 116 };

        // Application id PREFIX for fields generated by SuuntoPlus apps
        public static final byte[] SUUNTOPLUS_PREFIX = { -34, 19, -42, -88, -103, -61, -63, -99 };
    }

    enum AppType {
        UNKNOWN,
        SUUNTO_STANDARD,
        SUUNTOPLUS_METADATA,
        SUUNTOPLUS,
    }

    static class App {
        private final AppType type;
        private final String name;

        App(AppType type, String name) {
            this.type = type;
            this.name = name;
        }

        @Override
        public String toString() {
            return new StringJoiner(", ", App.class.getSimpleName() + "[", "]")
                    .add("type=" + type)
                    .add("name='" + name + "'")
                    .toString();
        }
    }

    static class Field {
        private final App app;
        private final String fieldName;
        private final String unit;
        private final String label;
        private final String format;

        Field(App app, String fieldName, String unit, String label, String format) {
            this.app = app;
            this.fieldName = fieldName;
            this.unit = unit;
            this.label = label;
            this.format = format;
        }

        @Override
        public String toString() {
            return new StringJoiner(", ", Field.class.getSimpleName() + "[", "]")
                    .add("app=" + app)
                    .add("fieldName='" + fieldName + "'")
                    .add("unit='" + unit + "'")
                    .add("label='" + label + "'")
                    .add("format='" + format + "'")
                    .toString();
        }
    }

    /**
     * Decodes given FIT file bytes and writes info on found developer fields into the log.
     * @param fitBytes The FIT file in byte array
     * @param log PrintStream where information is written, for example System.out.
     * @throws Exception if decoding fails or data is unexpected
     */
    public static void decode(byte[] fitBytes, PrintStream log) throws Exception {
        Decode decode = new Decode();
        MesgBroadcaster mesgBroadcaster = new MesgBroadcaster(decode);

        Map<Short, App> apps = new HashMap<>();
        Map<String, Field> fields = new HashMap<>();

        mesgBroadcaster.addListener((DeveloperDataIdMesgListener) (d) -> {
            AppType type = AppType.UNKNOWN;
            if (bytesEqual(d.getApplicationId(), ApplicationId.SUUNTO_STANDARD)) {
                type = AppType.SUUNTO_STANDARD;
            } else if (bytesEqual(d.getApplicationId(), ApplicationId.SUUNTOPLUS_METADATA)) {
                type = AppType.SUUNTOPLUS_METADATA;
            } else if (bytesStartWith(d.getApplicationId(), ApplicationId.SUUNTOPLUS_PREFIX)) {
                type = AppType.SUUNTOPLUS;
            }

            // The last 8 bytes of SuuntoPlus application id can be decoded into app name
            String name = null;
            if (type == AppType.SUUNTOPLUS) {
                ByteArrayOutputStream nameBaos = new ByteArrayOutputStream();
                Arrays.stream(d.getApplicationId()).skip(8).forEach(nameBaos::write);
                name = nameBaos.toString();
            }

            App app = new App(type, name);

            log.println("App " + d.getDeveloperDataIndex() + " = " + app);
            apps.put(d.getDeveloperDataIndex(), app);
        });

        mesgBroadcaster.addListener((FieldDescriptionMesgListener) (d) -> {
            String label = null;
            String format = null;

            for (DeveloperField df : d.getDeveloperFields()) {
                Field output = fields.get(df.getDeveloperDataIndex() + "_" + df.getNum());
                if (output != null && output.app.type == AppType.SUUNTOPLUS_METADATA) {
                    if ("suuntoplus_field_label".equals(output.fieldName)) {
                        label = df.getStringValue();
                    } else if ("suuntoplus_field_format".equals(output.fieldName)) {
                        format = df.getStringValue();
                    }
                }
            }

            App app = apps.get(d.getDeveloperDataIndex());

            String id = d.getDeveloperDataIndex() + "_" + d.getFieldDefinitionNumber();
            Field output = new Field(app, d.getFieldName(0), d.getUnits(0), label, format);

            log.println("Field " + id + " = " + output);
            fields.put(id, output);
        });

        mesgBroadcaster.addListener((SessionMesgListener) (d) -> {
            for (DeveloperField df : d.getDeveloperFields()) {
                Field output = fields.get(df.getDeveloperDataIndex() + "_" + df.getNum());
                log.println("Session " + output + " = " + df.getValue());
            }
        });

        try (InputStream byteStream = new ByteArrayInputStream(fitBytes)) {
            decode.read(byteStream, mesgBroadcaster, mesgBroadcaster);
        }
    }

    private static boolean bytesEqual(Byte[] source, byte[] match) {
        return source.length == match.length && bytesStartWith(source, match);
    }

    private static boolean bytesStartWith(Byte[] source, byte[] match) {
        if (match.length == 0 || match.length > source.length) {
            throw new IllegalArgumentException("Parameter match is too short or too long");
        }

        for (int i = 0; i < match.length; i++ ) {
            if (source[i] != match[i]) {
                return false;
            }
        }

        return true;
    }
}
